@@ -1,5 +1,5 @@
 from django import forms
-from django.contrib import admin, auth
+from django.contrib import admin, auth, sites
 from django.contrib.admin.views.main import ChangeList
 from django.db import connection
 from django.db.models import Count, F, Q
@@ -20,16 +20,18 @@ from urllib.parse import unquote, parse_qsl
 
 from ..forms import ProgramAdminForm, FormAdminForm, DependencyAdminForm, \
     StockBlockAdminForm, CustomBlockAdminForm, CollectionBlockAdminForm, \
-    SubmissionAdminForm, SubmissionItemAdminForm
+    SubmissionAdminForm, SubmissionItemAdminForm, SiteAdminForm, \
+    UserCreationAdminForm
 from ..models import Program, Form, FormLabel, FormBlock, FormDependency, \
-    CustomBlock, CollectionBlock, SubmissionRecord
+    CustomBlock, CollectionBlock, SubmissionRecord, Site
 from ..filetype import FileType
 from ..plugins import get_matching_plugin
 from ..signals import register_program_settings, register_form_settings, \
-    register_user_actions, form_published_changed
-from ..utils import submission_link
+    register_user_actions, form_published_changed, form_settings_changed
+from ..tasks import timed_complete_form
+from ..utils import submission_link, get_current_site, user_programs
 from .actions import UserActionsMixin, FormActionsMixin,FormBlockActionsMixin, \
-    SubmissionActionsMixin
+    SubmissionActionsMixin, download_view
 
 
 class FormativeAdminSite(admin.AdminSite):
@@ -52,6 +54,13 @@ class FormativeAdminSite(admin.AdminSite):
                     self.submissions_registered[form.item_model] = True
         
         form_published_changed.send(self)
+    
+    def get_urls(self):
+        urls = super().get_urls()
+        url = path('files_download/<int:form_id>/',
+                   self.admin_view(download_view),
+                   name='formative_files_download')
+        return [url] + urls
 
 
 site = FormativeAdminSite()
@@ -59,10 +68,97 @@ site = FormativeAdminSite()
 site.register(auth.models.Group, auth.admin.GroupAdmin)
 
 
-@admin.register(auth.models.User, site=site)
+@admin.register(auth.get_user_model(), site=site)
 class UserAdmin(UserActionsMixin, auth.admin.UserAdmin):
     change_list_template = 'admin/formative/user/change_list.html'
     actions = ['make_active', 'make_inactive', 'send_password_reset']
+    add_form = UserCreationAdminForm
+    add_fieldsets = ((None, {'classes': ('wide',),
+                             'fields': ('email', 'password1', 'password2',
+                                        'is_staff')}),)
+    
+    def get_queryset(self, request):
+        queryset, user = super().get_queryset(request), request.user
+        
+        if user.site: queryset = queryset.filter(site=user.site)
+        if not user.is_superuser:
+            queryset = queryset.filter(Q(is_staff=False) | Q(pk=user.pk))
+        return queryset
+    
+    def get_fieldsets(self, request, obj=None):
+        fieldsets = super().get_fieldsets(request, obj)
+        if not obj: return fieldsets
+        
+        if obj.is_staff:
+            main = ((None, {'fields': ('username', 'password', 'programs')}),)
+        else: main = ((None, {'fields': ('email', 'password')}),)
+        if request.user.is_superuser and not request.user.site:
+            main = ((None, {'fields': main[0][1]['fields'] + ('site',)}),)
+        
+        info = ((fieldsets[1][0], {
+            'fields': tuple(f for f in fieldsets[1][1]['fields']
+                            if obj.is_staff or f != 'email')
+        }),)
+        return main + info + fieldsets[2:]
+    
+    def get_list_display(self, request):
+        display = super().get_list_display(request)
+        keep = tuple(f for f in display if f not in ('email', 'username'))
+        if not request.user.is_superuser or request.user.site:
+            return ('user_id',) + keep + ('is_active', 'date_joined')
+        return ('user_id',) + keep + ('is_active', 'date_joined', 'site')
+    
+    def get_list_filter(self, request):
+        filters = super().get_list_filter(request)
+        if request.user.is_superuser and not request.user.site:
+            return filters + ('site',)
+        return filters
+    
+    def get_form(self, request, obj=None, **kwargs):
+        form = super().get_form(request, obj, **kwargs)
+        if not obj or not obj.is_staff:
+            form.base_fields['email'].required = True
+        
+        if 'programs' not in form.base_fields: return form
+        qs = form.base_fields['programs'].queryset
+        site = get_current_site(request)
+        if site: form.base_fields['programs'].queryset = qs.filter(sites=site)
+        return form
+    
+    def has_change_permission(self, request, obj=None):
+        if not request.user.is_superuser and obj and obj.is_staff:
+            if obj.pk != request.user.pk: return False
+        return super().has_change_permission(request, obj)
+    
+    def has_delete_permission(self, request, obj=None):
+        if not request.user.is_superuser and obj and obj.is_staff: return False
+        return super().has_delete_permission(request, obj)
+    
+    def get_readonly_fields(self, request, obj=None):
+        fields = super().get_readonly_fields(request, obj)
+        if not request.user.is_superuser and obj:
+            ro = ('programs', 'is_staff', 'is_superuser',
+                  'groups', 'user_permissions')
+            if obj.is_staff: ro = ('password',) + ro
+            return ro + fields
+        return fields
+    
+    def save_model(self, request, obj, form, change):
+        # TODO need a username form field that strips out __{obj.site_id}
+        site = get_current_site(request)
+        if not obj.is_staff or not request.user.is_superuser: obj.site = site
+        if request.user.is_superuser and request.user.site:
+            obj.site = request.user.site
+        # currently, this doesn't keep it updated when ID set, then changes
+        if not obj.is_staff: obj.username = obj.email
+        if obj.site and '__' not in obj.username:
+            obj.username += f'__{obj.site_id}'
+        super().save_model(request, obj, form, change)
+    
+    @admin.display(description='username')
+    def user_id(self, obj):
+        if obj.is_staff: return obj.username
+        return obj.email # we hide the __{site_id} username from regular admins
     
     def get_actions(self, request):
         actions = super().get_actions(request)
@@ -93,7 +189,11 @@ class ProgramAdmin(admin.ModelAdmin, DynamicArrayMixin):
     def get_fieldsets(self, request, obj=None):
         fields = self.get_fields(request, obj)
         
-        main = (None, {'fields': fields})
+        exclude = []
+        if not request.user.is_superuser or request.user.site:
+            exclude.append('sites')
+        
+        main = (None, {'fields': [ f for f in fields if f not in exclude ] })
         if not obj: return [main]
         
         responses = register_program_settings.send(self)
@@ -108,6 +208,26 @@ class ProgramAdmin(admin.ModelAdmin, DynamicArrayMixin):
         fields = super().get_readonly_fields(request, obj)
         if obj: fields += ('slug',)
         return fields
+    
+    def get_queryset(self, request):
+        queryset = super().get_queryset(request)
+        site = get_current_site(request)
+        if not request.user.is_superuser and not request.user.site:
+            queryset = queryset.filter(sites=site)
+        return user_programs(queryset, '', request)
+    
+    def get_changeform_initial_data(self, request):
+        initial = super().get_changeform_initial_data(request)
+        site = get_current_site(request)
+        if site and request.user.is_superuser:
+            initial['sites'] = (get_current_site(request),)
+        return initial
+    
+    def save_model(self, request, obj, form, change):
+        super().save_model(request, obj, form, change)
+        if not change and not request.user.is_superuser:
+            site = get_current_site(request)
+            if site: obj.sites.add(site)
 
 
 class FormChangeList(ChangeList):
@@ -124,7 +244,7 @@ class FormAdmin(FormActionsMixin, admin.ModelAdmin):
     list_display = ('name', 'program', 'created', 'modified')
     list_filter = ('program',)
     form = FormAdminForm
-    actions = ['form_plugins']
+    actions = ['form_plugins', 'export_json', 'duplicate']
     
     def get_changelist(self, request, **kwargs):
         return FormChangeList
@@ -159,6 +279,23 @@ class FormAdmin(FormActionsMixin, admin.ModelAdmin):
             else: fields += ('program', 'slug')
         return fields
     
+    def get_queryset(self, request):
+        queryset = super().get_queryset(request)
+        site = get_current_site(request)
+        if site: queryset = queryset.filter(program__sites=site)
+        return user_programs(queryset, 'program__', request)
+    
+    def get_form(self, request, obj=None, **kwargs):
+        form = super().get_form(request, obj, **kwargs)
+        
+        if 'program' not in form.base_fields: return form
+        qs = form.base_fields['program'].queryset
+        site = get_current_site(request)
+        if site:
+            qs = user_programs(qs.filter(sites=site), '', request)
+            form.base_fields['program'].queryset = qs
+        return form
+    
     def save_form(self, request, form, change):
         obj = super().save_form(request, form, change)
         if not change: return obj
@@ -184,8 +321,16 @@ class FormAdmin(FormActionsMixin, admin.ModelAdmin):
             if 'emails' in obj.options and not obj.options['emails']:
                 del obj.options['emails']
         
-        return obj
+        if 'timed_completion' in form.changed_data:
+            val = form.cleaned_data['timed_completion']
+            timed_complete_form.apply_async(args=(obj.id, val), eta=val)
         
+        return obj
+    
+    def save_model(self, request, obj, form, change):
+        super().save_model(request, obj, form, change)
+        form_settings_changed.send(obj, changed_data=form.changed_data)
+    
     def response_post_save_change(self, request, obj):
         app_label = self.model._meta.app_label
         url = reverse('admin:%s_formblock_formlist' % (app_label,),
@@ -212,6 +357,21 @@ class FormLabelAdmin(admin.ModelAdmin):
             attrs['class'], attrs['cols'] = 'vTextArea', 60
             formfield.widget = forms.Textarea(attrs=attrs)
         return formfield
+    
+    def get_form(self, request, obj=None, **kwargs):
+        form = super().get_form(request, obj, **kwargs)
+        
+        site = get_current_site(request)
+        qs = form.base_fields['form'].queryset.filter(program__sites=site)
+        form.base_fields['form'].queryset = user_programs(qs, 'program__',
+                                                          request)
+        return form
+    
+    def get_queryset(self, request):
+        queryset = super().get_queryset(request)
+        site = get_current_site(request)
+        return user_programs(queryset.filter(form__program__sites=site),
+                             'form__program__', request)
 
 
 class FormDependencyFormSet(forms.BaseInlineFormSet):
@@ -236,10 +396,10 @@ class FormDependencyInline(admin.TabularInline):
         
         return obj.form.status == Form.Status.DRAFT
     
-    def has_change_permission(self, request, obj):
+    def has_change_permission(self, request, obj=None):
         return self.has_add_permission(request, obj)
     
-    def has_delete_permission(self, request, obj):
+    def has_delete_permission(self, request, obj=None):
         return self.has_add_permission(request, obj)
 
 
@@ -276,7 +436,6 @@ class FormBlockBase:
         if form_id: form.form_id = form_id
         
         if obj and obj.form.status == Form.Status.DRAFT:
-            # TODO: where would we modify form.fields instead?
             if 'dependence' in form.base_fields:
                 qs = form.base_fields['dependence'].queryset
                 qs = qs.filter(form=obj.form, page__gt=0)
@@ -305,6 +464,12 @@ class FormBlockBase:
             if not isinstance(inline, FormDependencyInline) or obj is not None:
                 yield inline.get_formset(request, obj), inline
     
+    def get_queryset(self, request):
+        qs = super().get_queryset(request)
+        site = get_current_site(request)
+        return user_programs(qs.filter(form__program__sites=site),
+                             'form__program__', request)
+        
     def response_add(self, request, obj, **kwargs):
         if '_popup' in request.POST or '_continue' not in request.POST:
             return super().response_add(request, obj, **kwargs)
@@ -509,6 +674,7 @@ class FormBlockAdmin(FormBlockActionsMixin, FormBlockBase,
     
     def get_queryset(self, request):
         qs = super().get_queryset(request)
+        
         match, app_label = request.resolver_match, self.model._meta.app_label
         if match and match.url_name == f'{app_label}_formblock_formlist':
             qs = qs.filter(form_id=match.kwargs['form_id'])
@@ -529,6 +695,10 @@ class FormBlockAdmin(FormBlockActionsMixin, FormBlockBase,
         else: kwargs['widgets'] = {'_rank': forms.HiddenInput}
         return super().get_changelist_form(request, **kwargs)
     
+    def add_view(self, *args, **kwargs):
+        kwargs['extra_context'] = {'show_save_and_add_another': False}
+        return super().add_view(*args, **kwargs)
+    
     def change_view(self, *args, **kwargs):
         # we still have the bulk action for delete - it redirects properly
         kwargs['extra_context'] = {'show_delete': False,
@@ -653,6 +823,32 @@ class CollectionBlockAdmin(FormBlockChildAdmin, DynamicArrayMixin):
         return fields
 
 
+class SuperuserAccessMixin:
+    def has_module_permission(self, request): return request.user.is_superuser
+    
+    def has_view_permission(self, request, obj=None):
+        return request.user.is_superuser
+    def has_change_permission(self, request, obj=None):
+        return request.user.is_superuser
+    def has_delete_permission(self, request, obj=None):
+        return request.user.is_superuser
+
+class SiteAccessMixin:
+    def has_change_permission(self, request, obj=None):
+        slug = self.model._meta.program_slug
+        site = get_current_site(request)
+        
+        programs = user_programs(site.programs, '', request)
+        return slug in programs.values_list('db_slug', flat=True)
+    
+    def has_view_permission(self, request, obj=None):
+        return self.has_change_permission(request, obj)
+    def has_delete_permission(self, request, obj=None):
+        return self.has_change_permission(request, obj)
+    def has_add_permission(self, request):
+        return self.has_change_permission(request, None)
+
+
 class SubmittedListFilter(admin.SimpleListFilter):
     title = 'status'
     parameter_name = '_submitted'
@@ -675,7 +871,7 @@ class SubmissionRecordInline(admin.TabularInline):
     exclude = ('program', 'form', 'submission')
     readonly_fields = ('type', 'recorded', 'text', 'number', 'deleted')
     
-    def has_add_permission(self, request, obj=None): return False
+    def has_add_permission(self, request, obj): return False
     
     def get_formset(self, request, obj=None, **kwargs):
         return forms.modelformset_factory(self.model, **{
@@ -686,13 +882,21 @@ class SubmissionRecordInline(admin.TabularInline):
         })
 
 
-class SubmissionAdmin(SubmissionActionsMixin, admin.ModelAdmin):
+class SubmissionAdmin(SiteAccessMixin, SubmissionActionsMixin,
+                      admin.ModelAdmin):
     list_display = ('_email', '_created', '_modified', '_submitted')
     list_filter = ('_email', SubmittedListFilter)
     readonly_fields = ('_submitted', 'items_index',)
     form = SubmissionAdminForm
     inlines = [SubmissionRecordInline]
-    actions = ['send_email', 'export_csv']
+    actions = ['send_email', 'export_csv', 'download_files']
+    
+    def get_actions(self, request):
+        actions = super().get_actions(request)
+        desc = 'Delete selected submissions'
+        if 'delete_selected' not in actions: return actions
+        actions['delete_selected'][0].short_description = desc
+        return actions
     
     def delete_queryset(self, request, queryset):
         # something is up with model registry. manually delete the related items
@@ -729,7 +933,7 @@ class SubmissionAdmin(SubmissionActionsMixin, admin.ModelAdmin):
         return url
 
 
-class SubmissionItemAdmin(admin.ModelAdmin):
+class SubmissionItemAdmin(SiteAccessMixin, admin.ModelAdmin):
     list_display = ('_id', '_submission', '_collection', '_rank', '_file')
     list_filter = (
         '_submission', '_collection',
@@ -741,3 +945,23 @@ class SubmissionItemAdmin(admin.ModelAdmin):
     
     def has_add_permission(self, request):
         return False
+
+
+@admin.register(Site, site=site)
+class SiteAdmin(SuperuserAccessMixin, admin.ModelAdmin):
+    form = SiteAdminForm
+    list_display = ('domain', 'name', 'time_zone')
+    search_fields = ('domain', 'name')
+    
+    def get_queryset(self, request):
+        queryset = super().get_queryset(request)
+        if request.user.site: return queryset.filter(pk=request.user.site.pk)
+        return queryset
+    
+    def save_model(self, request, obj, form, change):
+        super().save_model(request, obj, form, change)
+        Site.objects.clear_cache()
+    
+    def delete_model(self, request, obj):
+        super().delete_model(request, obj)
+        Site.objects.clear_cache()
