@@ -3,7 +3,8 @@ __all__ = ["OpenForm", "SubmissionForm", "ItemFileForm", "ItemsForm",
            "ItemsFormSet", "AdminJSONForm", "ProgramAdminForm", "FormAdminForm",
            "StockBlockAdminForm", "CustomBlockAdminForm",
            "CollectionBlockAdminForm", "SubmissionAdminForm",
-           "SubmissionItemAdminForm", "EmailAdminForm"]
+           "SubmissionItemAdminForm", "EmailAdminForm", "SiteAdminForm",
+           "UserCreationAdminForm"]
 
 from .formative import *
 from .admin import *
