@@ -1,17 +1,19 @@
 from django import forms
 from django.core.exceptions import ValidationError
 from django.contrib.admin import widgets
+from django.contrib.auth.forms import UserCreationForm
 from copy import deepcopy
 from django_better_admin_arrayfield.forms.fields import DynamicArrayField
 from django_better_admin_arrayfield.forms.widgets import DynamicArrayWidget
 
-import datetime
+from datetime import datetime
+import pytz, babel.dates
 
 from ..signals import register_program_settings, register_form_settings
 from ..filetype import FileType
 from ..stock import StockWidget
 from ..models import Program, Form, FormBlock, CustomBlock, CollectionBlock, \
-    Submission, SubmissionItem
+    Submission, SubmissionItem, Site, User
 from ..plugins import get_available_plugins
 from ..validators import validate_program_identifier, validate_form_identifier,\
     validate_formblock_identifier
@@ -164,7 +166,7 @@ class JSONDateTimeWidget(widgets.AdminSplitDateTime):
     def decompress(self, val):
         if not val: return super().decompress(val)
         
-        return super().decompress(datetime.datetime.fromisoformat(val))
+        return super().decompress(datetime.fromisoformat(val))
 
 
 class JSONDateTimeField(forms.SplitDateTimeField):
@@ -303,6 +305,17 @@ class AdminJSONForm(forms.ModelForm, metaclass=AdminJSONFormMetaclass):
         return cleaned_data
 
 
+class UserCreationAdminForm(UserCreationForm):
+    class Meta:
+        fields = ('email', 'is_staff')
+        model = User
+    
+    def save(self, commit=True):
+        user = super().save(commit=False)
+        user.username = user.email
+        if commit: user.save()
+        return user
+
 class ProgramAdminForm(AdminJSONForm):
     name = forms.CharField(
         max_length=Program._meta.get_field('name').max_length,
@@ -322,11 +335,13 @@ class ProgramAdminForm(AdminJSONForm):
     )
     
     class Meta:
-        static_fields = ('name', 'slug', 'description', 'hidden', 'home_url')
+        static_fields = ('name', 'slug', 'description', 'hidden', 'home_url',
+                         'sites')
         json_fields = {'options': ['home_url']}
         dynamic_fields = True
         widgets = {
-            'description': widgets.AdminTextareaWidget(attrs={'rows': 3})
+            'description': widgets.AdminTextareaWidget(attrs={'rows': 3}),
+            'sites': forms.CheckboxSelectMultiple()
         }
     
     def __init__(self, *args, **kwargs):
@@ -337,6 +352,8 @@ class ProgramAdminForm(AdminJSONForm):
             kwargs['admin_fields'] = admin_fields
         
         super().__init__(*args, **kwargs)
+        if 'sites' in self.fields:
+            self.fields['sites'].widget.can_add_related = False
 
 
 class FormAdminForm(AdminJSONForm):
@@ -357,6 +374,18 @@ class FormAdminForm(AdminJSONForm):
         widget=widgets.AdminRadioSelect(attrs={'class': 'radiolist'})
     )
     hidden = forms.BooleanField(required=False)
+    enabled_message = forms.CharField(required=False,
+        widget=widgets.AdminTextInputWidget,
+        help_text='If not specified, default is "Open for submissions."'
+    )
+    disabled_message = forms.CharField(required=False,
+        widget=widgets.AdminTextInputWidget,
+        help_text='If not specified, default is "Not yet open for submissions."'
+    )
+    completed_message = forms.CharField(required=False,
+        widget=widgets.AdminTextInputWidget,
+        help_text='If not specified, default is "Closed."'
+    )
     access_enable = forms.CharField(required=False, label='?access= password')
     review_pre = forms.CharField(
         required=False, widget=widgets.AdminTextareaWidget(attrs={'rows': 5}),
@@ -390,7 +419,8 @@ class FormAdminForm(AdminJSONForm):
     class Meta:
         static_fields = ('program', 'name', 'slug', 'status', 'hidden')
         json_fields = {'options': [
-            'hidden', 'access_enable', 'review_pre', 'review_post',
+            'hidden', 'enabled_message', 'disabled_message',
+            'completed_message', 'access_enable', 'review_pre', 'review_post',
             'submitted_review_pre', 'timed_completion', 'complete_submit_time',
             'no_review_after_submit', 'thanks', 'emails'
         ]}
@@ -772,6 +802,27 @@ class SubmissionItemAdminForm(forms.ModelForm):
             if field.required: field.required = False
 
 
+offsets = {}
+for name in pytz.common_timezones:
+    zone = pytz.timezone(name)
+    offset = int(zone.utcoffset(datetime(2022, 7, 1)).total_seconds()) // 60
+    if offset not in offsets: offsets[offset] = []
+    display_name = babel.dates.get_timezone_name(babel.dates.get_timezone(name))
+    offsets[offset].append((name, f'{display_name} - {name}'))
+
+TZ_CHOICES = [ (('+' if k >= 0 else '') + f'{k // 60}:{k % 60:02d}',
+                sorted(offsets[k], key=lambda x: x[1]))
+               for k in sorted(offsets.keys()) ]
+
+class SiteAdminForm(forms.ModelForm):
+    time_zone = forms.ChoiceField(choices=[('', '[not localized]')]+TZ_CHOICES)
+    
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.fields['time_zone'].widget.attrs['data-dropdown-css-class'] = \
+            'larger'
+
+
 class EmailAdminForm(forms.Form):
     name = forms.ChoiceField(choices=(), initial='confirmation')
     subject = forms.CharField(widget=widgets.AdminTextInputWidget)
