@@ -13,6 +13,7 @@ from django.utils.translation import gettext_lazy as _
 from django.utils import timezone
 from django.urls import reverse
 from polymorphic.models import PolymorphicModel
+from polymorphic.managers import PolymorphicManager
 import uuid
 from itertools import groupby
 from pathlib import Path
@@ -30,10 +31,16 @@ from .automatic import AutoSlugModel
 markdown = MarkdownFormatter()
 
 
+class ProgramManager(models.Manager):
+    def get_by_natural_key(self, slug):
+        return self.get(slug=slug)
+
 class Program(AutoSlugModel):
     class Meta:
         ordering = ['created']
     
+    sites = models.ManyToManyField('Site', blank=True, related_name='programs',
+                                   related_query_name='program')
     name = models.CharField(max_length=64)
     slug = models.SlugField(max_length=30, unique=True, allow_unicode=True,
                             verbose_name='identifier')
@@ -43,10 +50,15 @@ class Program(AutoSlugModel):
     options = models.JSONField(default=dict, blank=True)
     hidden = models.BooleanField(default=False)
     created = models.DateTimeField(auto_now_add=True)
+    
+    objects = ProgramManager()
 
     def __str__(self):
         return self.name
     
+    def natural_key(self):
+        return (self.slug,)
+    
     def validate_unique(self, exclude=None):
         super().validate_unique(exclude=exclude)
         if not self._state.adding: return
@@ -64,6 +76,10 @@ class Program(AutoSlugModel):
         return None
 
 
+class FormManager(models.Manager):
+    def get_by_natural_key(self, program_slug, slug):
+        return self.get(program__slug=program_slug, slug=slug)
+
 class Form(AutoSlugModel):
     class Meta:
         constraints = [
@@ -99,10 +115,15 @@ class Form(AutoSlugModel):
     validation_type = models.CharField(max_length=16, editable=False,
                                        default=Validation.EMAIL,
                                        choices=Validation.choices)
+    objects = FormManager()
     
     def __str__(self):
         return self.name
     
+    def natural_key(self):
+        return self.program.natural_key() + (self.slug,)
+    natural_key.dependencies = ['formative.program']
+    
     def validate_unique(self, exclude=None):
         super().validate_unique(exclude=exclude)
         if not self._state.adding: return
@@ -329,9 +350,15 @@ class Form(AutoSlugModel):
         if self.status == self.Status.DRAFT:
             return 'NA'
         elif self.status == self.Status.DISABLED:
+            if 'disabled_message' in self.options:
+                return self.options['disabled_message']
             return _('Not yet open for submissions')
         elif self.status == self.Status.COMPLETED:
+            if 'completed_message' in self.options:
+                return self.options['completed_message']
             return _('Closed')
+        if 'enabled_message' in self.options:
+            return self.options['enabled_message']
         return _('Open for submissions')
     
     def hidden(self):
@@ -424,6 +451,11 @@ class Form(AutoSlugModel):
         return emails
 
 
+class FormLabelManager(models.Manager):
+    def get_by_natural_key(self, program_slug, form_slug, path, style):
+        return self.get(form__program__slug=program_slug, form__slug=form_slug,
+                        path=path, style=style)
+
 class FormLabel(models.Model):
     class Meta:
         constraints = [
@@ -443,9 +475,15 @@ class FormLabel(models.Model):
     style = models.CharField(max_length=16, choices=LabelStyle.choices,
                              default=LabelStyle.WIDGET)
     
+    objects = FormLabelManager()
+    
     def __str__(self):
         return self.path
     
+    def natural_key(self):
+        return self.form.natural_key() + (self.path, self.style)
+    natural_key.dependencies = ['formative.form']
+    
     def display(self, inline=False):
         s = markdown.convert(self.text)
         if inline: return mark_safe(remove_p(s))
@@ -455,6 +493,12 @@ class FormLabel(models.Model):
         return self.display(inline=True)
 
 
+class FormDependencyManager(models.Manager):
+    def get_by_natural_key(self, program_slug, form_slug, name, value):
+        return self.get(block__form__program__slug=program_slug,
+                        block__form__slug=form_slug,
+                        block__name=name, value=value)
+
 class FormDependency(models.Model):
     class Meta:
         verbose_name = 'dependency value'
@@ -468,12 +512,24 @@ class FormDependency(models.Model):
                               related_query_name='dependency')
     value = models.CharField(max_length=64, blank=True)
     
+    objects = FormDependencyManager()
+    
     def __str__(self):
         if self.block.dependence:
             return f'{self.block.dependence.name}="{self.value}"'
         return f'?="{self.value}"'
+    
+    def natural_key(self):
+        return self.block.natural_key() + (self.value,)
+    natural_key.dependencies = ['formative.formblock', 'formative.customblock',
+                                'formative.collectionblock']
 
 
+class FormBlockManager(PolymorphicManager):
+    def get_by_natural_key(self, program_slug, form_slug, name):
+        return self.non_polymorphic().get(form__program__slug=program_slug,
+                                          form__slug=form_slug, name=name)
+
 class FormBlock(PolymorphicModel, RankedModel):
     class Meta(PolymorphicModel.Meta, RankedModel.Meta):
         constraints = [
@@ -494,10 +550,15 @@ class FormBlock(PolymorphicModel, RankedModel):
                                    related_query_name='dependent')
     negate_dependencies = models.BooleanField(default=False,
                                               verbose_name='negate dependency')
+    objects = FormBlockManager()
     
     def __str__(self):
         return self.name
     
+    def natural_key(self):
+        return self.form.natural_key() + (self.name,)
+    natural_key.dependencies = ['formative.form']
+    
     def rank_group(self):
         return FormBlock.objects.filter(form=self.form, page=self.page)
     
